/////////////////////////////////////////////////////////
// IOModel_X.cpp
// load/save Frame descriptions
//
/////////////////////////////////////////////////////////

#include "ToolBox\IOModel_X.h"

//#include <float> //see Float_Anomaly.txt

#define TEXT_BUFFER 255

#define MAX_TEMPLATES 15

struct XOF_TEMPLATEID{
	char* TxtID;
	uint16 TemplateID;
};

XOF_TEMPLATEID Templates[MAX_TEMPLATES] = {
	{"template" ,X_TEMPLATE},
   {"FrameTransformMatrix", X_FRAMETRANSFORMMATRIX},
   {"Frame", X_FRAME},
   {"XSkinMeshHeader", X_SKINMESHHEADER},
   {"MeshTextureCoords", X_MESHTEXTURECOORDS},
	{"MeshMaterialList", X_MESHMATERIALLIST},
   {"MeshNormals", X_MESHNORMALS},
   {"Mesh", X_MESH},
   {"Material", X_MATERIAL},
   {"SkinWeights", X_SKINWEIGHTS},
   {"TextureFilename", X_TEXTUREFILENAME},
   {"AnimationSet", X_ANIMATIONSET},
   {"AnimationKey", X_ANIMATIONKEY},
   {"Animation", X_ANIMATION},
	{"Header", X_HEADER}
};

//////////////////////////////////////////////////////////
//
//       MAIN LOAD & SAVE FUNCTIONS
//
//////////////////////////////////////////////////////////

bool ToolBox::IO_Model_X::Load(std::string pFilename, Frm::Model3D* &pT)
{
	XFileHeader XHeader;

   MYTRACE("Processing file:", pFilename);

	//Open the file
	fin.open(pFilename.c_str(), ios::in);

   if (fin.bad())
   	return false;

   //Check the file Header
   fin.read((char*)&XHeader,16);
   if (XHeader.Magic != XOFFILE_FORMAT_MAGIC)
   	{
      MYTRACE("Not a .X model file. Aborted...");
   	return false;
      }

	if (XHeader.Major_Version != XOFFILE_FORMAT_VERSION03)
   	{
      MYTRACE("Major version greater than 03. Aborted...");
   	return false;
      }

   if ((XHeader.Minor_Version != XOFFILE_FORMAT_VERSION03) || (XHeader.Minor_Version != XOFFILE_FORMAT_VERSION02))
   	{
      MYTRACE("Minor version greater than 03. Aborted...");
		return false;
      }

	if (XHeader.Format != XOFFILE_FORMAT_TEXT)
   	{
      MYTRACE("Not a text format. Aborted...");
   	return false;
      }

   _Object = pT;

   //Enter the main processing loop
   while(!fin.eof())
   	{
      //Read the block Name
      //If it is a Mesh we process it.
      //If it is a Frame we process it.
      //Else we avoid that block and go to the next one
      switch (ProcessBlock()) {
      	case X_ERROR: MYTRACE("Stopped processing the file ..."); return false;
         case X_COMMENT: break; //nothing to do
         case X_EBRACE: break; //end of a block ?!
         case X_FRAME: ProcessBone(); break;
         case X_MESH: ProcessMesh(); break;
         case X_OBRACE:
         case X_UNKNOWN:
         default: AvoidTemplate(); break;
      	}
      }

   MYTRACE("Processed file:", pFilename, "OK");

   fin.close();
   return true;
}

bool ToolBox::IO_Model_X::Save(std::string pFilename, Frm::Model3D* &pT)
{
	return false;
}

//////////////////////////////////////////////////////////
//
//       UTILITIES
//
//////////////////////////////////////////////////////////

int16 ToolBox::IO_Model_X::ProcessBlock(void)
{
	std::string Text;
	char Token = fin.peek();
   switch (Token) {
   	case '\n':
   	case ' ':
      case '\t': fin.get(); return X_COMMENT; //spaces are identified as comments
      case '{': return X_OBRACE;
      case '}': fin.get(); return X_EBRACE; //We arrived at the end of the block
      case '/': fin.get(); if (fin.peek() != '/'){return X_ERROR;} //we processed a block name starting with slash ?!
      case '#': fin.ignore(TEXT_BUFFER, '\n'); return X_COMMENT;
      default:
      	//valid character, let's check if we recognize the block
      	fin >> Text;
         return BlockID(Text);
   	};
}

int16 ToolBox::IO_Model_X::BlockID(std::string &pText)
{
   long Pos;

   if (fin.eof())
   	return X_COMMENT;

   if (pText.empty())
   	{
      MYTRACE("Error, no block read !");
   	return X_ERROR;
      }

   //BEWARE: when using STL to compare text strings
   //always sort the list of strings to compare from
   //the longest to the shortest.
   //For STL String, FrameTransformMatrix matches both
   //FrameTransformMatrix and Frame !
   for(int i=0; i < MAX_TEMPLATES; i++)
   	{
		Pos = pText.find(Templates[i].TxtID);
   	if (Pos > -1)
      	{
   		return Templates[i].TemplateID;
         }
		}
   MYTRACE("Unknown Block:", pText);
   return X_UNKNOWN;
}

void ToolBox::IO_Model_X::AvoidTemplate(void){
	char Token;

   fin.ignore(TEXT_BUFFER, '{');

	while (!fin.eof()){
   	Token = fin.peek();
      if (Token == '{')
      	AvoidTemplate();
      if	(Token == '}')
      	{
         fin.get();
      	return;
         }
      fin.get();
   	}
}

void ToolBox::IO_Model_X::Find(uchar pChar)
{
	fin.ignore(TEXT_BUFFER, pChar);
}

char* ToolBox::IO_Model_X::SetUID(char pType)
{
	//This is a quick hack to derive a Unique ID for blocks with
   //no identifier names like in the tiny_4anim.x example.

	_X_UID.Integer = GetTickCount(); //This function return a 4 byte wide number
   _X_UID.Text[4] = pType; //We set the 5th byte with a significant character

   //If any of the first 4 bytes are under 32 we add 32
   //We want to avoid the occurrence of the char '\0' within
   //the first 4 bytes since this would truncate the text returned.
   if (_X_UID.Text[0] < 32) _X_UID.Text[0]+=32;
   if (_X_UID.Text[1] < 32) _X_UID.Text[1]+=32;
   if (_X_UID.Text[2] < 32) _X_UID.Text[2]+=32;
   if (_X_UID.Text[3] < 32) _X_UID.Text[3]+=32;
   return _X_UID.Text;
}

//////////////////////////////////////////////////////////
//
//       Finding Mesh in Bone
//
//////////////////////////////////////////////////////////

void ToolBox::IO_Model_X::ProcessBone(void)
{
   int16 Token;
   char Data[TEXT_BUFFER];

   Find('{');

   //Enter the processing loop
   Token = X_OBRACE;
   while(Token != X_EBRACE)
   	{
      //Read the block Name
      //If it is a Mesh we process it.
      //If it is a Frame we process it.
      //Else we avoid that block and go to the next one
      Token = ProcessBlock();
      switch (Token) {
      	case X_COMMENT: break; //used for spaces and other kind of comments
         case X_EBRACE: return; //this is the end, my only friend ...
         case X_OBRACE: fin.getline(Data, TEXT_BUFFER, '}'); break; //a mesh name has been found here
     		case X_FRAME: ProcessBone(); break;
         case X_MESH: ProcessMesh(); break;
         case X_UNKNOWN:
         default:
         	AvoidTemplate(); break;
   		}
      }
}

//////////////////////////////////////////////////////////
//
//       MESH
//
//////////////////////////////////////////////////////////

void ToolBox::IO_Model_X::ProcessMesh(void)
{
   std::string Text;
   int16 Token;
   char Data[TEXT_BUFFER];

   //Create an empty mesh and initialise it.
   _LoadMesh = new Frm::Mesh;
   if (!_Object->_Meshes.empty())
   	{
      //These elements are used for Mesh concatenation
      Frm::Mesh* LastMesh = _Object->_Meshes.back();
      _LoadMesh->_FirstVertex = LastMesh->_FirstVertex + LastMesh->_nVertices;
      _LoadMesh->_FirstFace = LastMesh->_FirstFace + LastMesh->_nFaces;
      _LoadMesh->_FirstTextureCoord = LastMesh->_FirstTextureCoord + LastMesh->_nTextureCoords;
      _LoadMesh->_FirstMaterial = LastMesh->_FirstMaterial + LastMesh->_nMaterials;
      if (_LoadMesh->_FirstTextureCoord < _LoadMesh->_FirstVertex)
      	_LoadMesh->_FirstTextureCoord = _LoadMesh->_FirstVertex;
      _LoadMesh->_FirstNormal = LastMesh->_FirstNormal + LastMesh->_nNormals;
      if (_LoadMesh->_FirstNormal < _LoadMesh->_FirstVertex)
      	_LoadMesh->_FirstNormal = _LoadMesh->_FirstVertex;
      MYTRACE("Starting Vertex index:", _LoadMesh->_FirstVertex);
      MYTRACE("Starting Face index:", _LoadMesh->_FirstFace);
      MYTRACE("Starting TextureCoord index:", _LoadMesh->_FirstTextureCoord);
      MYTRACE("Starting Normal index:", _LoadMesh->_FirstNormal);
      MYTRACE("Starting Material index:", _LoadMesh->_FirstMaterial);
      }

   //Read in the mesh name
   //If there is no name, assign a name to that mesh
   fin.get(); //eating the whitespace ...
   Token = fin.peek();
    
   if (Token != '{')
   	fin >> _LoadMesh->_Name;
   else
   	_LoadMesh->_Name = SetUID('M');

  	Find('{');
   MYTRACE("Mesh:", _LoadMesh->_Name);

   //Read in the number of vertices
   fin.getline(Data, TEXT_BUFFER, ';');
   _LoadMesh->_nVertices = (uint16)TextToNum(Data);
   MYTRACE("Number of vertices:", _LoadMesh->_nVertices);
   _LoadMesh->_Vertices = new Frm::Vertex[_LoadMesh->_nVertices];

   //Load in the list of vertices
   for(int i=0; i< _LoadMesh->_nVertices; i++)
   	{
      fin.getline(Data, TEXT_BUFFER, ';');
   	_LoadMesh->_Vertices[i].data[0] = TextToNum(Data);
      fin.getline(Data, TEXT_BUFFER, ';');
   	_LoadMesh->_Vertices[i].data[1] = TextToNum(Data);
      fin.getline(Data, TEXT_BUFFER, ';');
   	_LoadMesh->_Vertices[i].data[2] = TextToNum(Data);
      fin.get();//eats either the comma or the semicolon at the end of each vertex description
      }

   //Read in the number of faces
   fin.getline(Data, TEXT_BUFFER, ';');
   _LoadMesh->_nFaces = (uint16)TextToNum(Data);
   MYTRACE("Number of Faces:", _LoadMesh->_nFaces);
   _LoadMesh->_Faces = new Frm::Face[_LoadMesh->_nFaces];

   //Load in the list of faces
   for(int i=0; i< _LoadMesh->_nFaces; i++)
   	{
      Find(';');
      fin.getline(Data, TEXT_BUFFER, ',');
   	_LoadMesh->_Faces[i].data[0] = (uint16)TextToNum(Data);
      fin.getline(Data, TEXT_BUFFER, ',');
   	_LoadMesh->_Faces[i].data[1] = (uint16)TextToNum(Data);
      fin.getline(Data, TEXT_BUFFER, ';');
   	_LoadMesh->_Faces[i].data[2] = (uint16)TextToNum(Data);
      fin.get(); //eats either the comma or the semicolon at the end of each face description
      }

   //Enter the Mesh processing loop
   Token = X_COMMENT;
   while(Token != X_EBRACE)
   	{
      //While we have not reached the end of the block
      //Read the block name. If the block name is recognized,
      //process it else avoid.
		Token = ProcessBlock();
  		switch (Token) {
      	case X_COMMENT: break; //used for spaces and other kind of comments
         case X_EBRACE: _Object->_Meshes.push_back(_LoadMesh);return; //this is the end, my only friend ...
         case X_MESHNORMALS: ProcessMeshNormals(); break;
      	case X_MESHTEXTURECOORDS: ProcessMeshTextureCoords(); break;
         case X_MESHMATERIALLIST: ProcessMeshMaterials(); break;
         default:
         	AvoidTemplate(); break;
      	}
		}
	_Object->_Meshes.push_back(_LoadMesh);
}

//////////////////////////////////////////////////////////
//
//       TEXTURE COORDS
//
//////////////////////////////////////////////////////////

void ToolBox::IO_Model_X::ProcessMeshTextureCoords(void)
{
   char Data[TEXT_BUFFER];

  	Find('{');

   //Read in the number of texture coordinates
   fin.getline(Data, TEXT_BUFFER, ';');
   _LoadMesh->_nTextureCoords = (uint16)TextToNum(Data);
   MYTRACE("Number of Texture Coords:", _LoadMesh->_nTextureCoords);
   _LoadMesh->_TextureCoords = new Frm::TCoord[_LoadMesh->_nTextureCoords];

   //Load the list of texture coordinates within the mesh object
   for(int i=0; i< _LoadMesh->_nTextureCoords; i++)
   	{
      fin.getline(Data, TEXT_BUFFER, ';');
   	_LoadMesh->_TextureCoords[i].data[0] = TextToNum(Data);
      fin.getline(Data, TEXT_BUFFER, ';');
   	_LoadMesh->_TextureCoords[i].data[1] = TextToNum(Data);
      fin.get();//eats the comma or the semicolon at the end
      }
   Find('}');
}

//////////////////////////////////////////////////////////
//
//       MESH NORMAL VECTORS
//
//////////////////////////////////////////////////////////

void ToolBox::IO_Model_X::ProcessMeshNormals(void)
{
   char Data[TEXT_BUFFER];

  	Find('{');
   //Read in the number of vectors
   fin.getline(Data, TEXT_BUFFER, ';');
   _LoadMesh->_nNormals = (uint16)TextToNum(Data);
   MYTRACE("Number of normals :", _LoadMesh->_nNormals);
   _LoadMesh->_Normals = new Frm::vector<float>[_LoadMesh->_nNormals];
   //Load in the list of vectors
   for(int i=0; i< _LoadMesh->_nNormals; i++)
   	{
      fin.getline(Data, TEXT_BUFFER, ';');
   	_LoadMesh->_Normals[i].x = TextToNum(Data);
      fin.getline(Data, TEXT_BUFFER, ';');
   	_LoadMesh->_Normals[i].y = TextToNum(Data);
      fin.getline(Data, TEXT_BUFFER, ';');
   	_LoadMesh->_Normals[i].z = TextToNum(Data);
      fin.get();//eats the comma or the semicolon at the end
      }

   _LoadMesh->_FaceNormals = new Frm::Face[_LoadMesh->_nFaces];
   //Load in the list of vector indices per face
   for(int i=0; i< _LoadMesh->_nFaces; i++)
   	{
      Find(';');
      fin.getline(Data, TEXT_BUFFER, ',');
   	_LoadMesh->_FaceNormals[i].data[0] = (uint16)TextToNum(Data);
      fin.getline(Data, TEXT_BUFFER, ',');
   	_LoadMesh->_FaceNormals[i].data[1] = (uint16)TextToNum(Data);
      fin.getline(Data, TEXT_BUFFER, ';');
   	_LoadMesh->_FaceNormals[i].data[2] = (uint16)TextToNum(Data);
      fin.get(); //eats either the comma or the semicolon at the end of each face description
      }

   Find('}');
}


//////////////////////////////////////////////////////////
//
//       MATERIALS USED IN MESH
//
//////////////////////////////////////////////////////////

void ToolBox::IO_Model_X::ProcessMeshMaterials(void)
{
   std::string Text;
   int16 Token;
   char Data[TEXT_BUFFER];

   Find('{');

   //Read in the number of materials used by that mesh
   fin.getline(Data, TEXT_BUFFER, ';');
   _LoadMesh->_nMaterials= (uint16)TextToNum(Data);
   MYTRACE("Number of Materials:", (uint16)TextToNum(Data));

   //Load in the material index for each face.
   fin.getline(Data, TEXT_BUFFER, ';');
   _LoadMesh->_FaceMaterials = new uint16[(uint16)TextToNum(Data)];
   for(int i=0; i< _LoadMesh->_nFaces-1; i++)
   	{
      fin.getline(Data, TEXT_BUFFER, ',');
   	_LoadMesh->_FaceMaterials[i] = (uint16)TextToNum(Data);
      }
   fin.getline(Data, TEXT_BUFFER, ';');
 	_LoadMesh->_FaceMaterials[_LoadMesh->_nFaces - 1] = (uint16)TextToNum(Data);
   fin.get(); //eats the last semicolon

   //Enter the Material list processing loop
   Token = X_COMMENT;
   while(Token != X_EBRACE)
   	{
      //While we have not reached the end of the block
      //Read the block name. If the block name is recognized,
      //process it else avoid.
      Token = ProcessBlock();
  		switch (Token) {
      	case X_COMMENT: break; //used for spaces and other kind of comments
         case X_EBRACE: return; //this is the end, my only friend ...
   		case X_MATERIAL: ProcessMaterial(); break;
         default:
         	AvoidTemplate(); break;
			}
		}
}

//////////////////////////////////////////////////////////
//
//       MATERIAL DESCRIPTION
//
//////////////////////////////////////////////////////////

void ToolBox::IO_Model_X::ProcessMaterial(void)
{
   std::string Text;
   int16 Token;
   char Data[TEXT_BUFFER];

   Frm::Material* NewMaterial = new Frm::Material;

   Find('{');
   for(int i=0; i < 4; i++)
   	{
   	fin.getline(Data, TEXT_BUFFER, ';');
   	NewMaterial->_FaceColor.data[i] = TextToNum(Data);
      }
   fin.get(); //eats the last semicolon
   fin.getline(Data, TEXT_BUFFER, ';');
   NewMaterial->_power = TextToNum(Data);

   for(int i=0; i < 3; i++)
   	{
   	fin.getline(Data, TEXT_BUFFER, ';');
   	NewMaterial->_SpecularColor.data[i] = TextToNum(Data);
      }
   fin.get();//eats the last semicolon

   for(int i=0; i < 3; i++)
   	{
   	fin.getline(Data, TEXT_BUFFER, ';');
   	NewMaterial->_EmissiveColor.data[i] = TextToNum(Data);
      }
   fin.get();//eats the last semicolon

   Token = X_COMMENT;
   while(Token != '}')
   	{
      Token = ProcessBlock();
  		switch (Token) {
      	case X_COMMENT: break; //used for spaces and other kind of comments
         case X_EBRACE: _LoadMesh->_Materials.push_back(NewMaterial);return; //this is the end, my only friend ...
      	case X_TEXTUREFILENAME:
         	Find('{');
            Find('"');
            //Read in the bitmap name
            fin.getline(Data, TEXT_BUFFER, '"');
            NewMaterial->_TextureName = Data;
            Find('}');
            break;
         default:
         	AvoidTemplate(); break;
         }
		}
   _LoadMesh->_Materials.push_back(NewMaterial);
}


